unit Otello;

{ Ingemar Ragnemalm as converted a C otello to Pascal
  this is my Delphi component version of Otello.

  Note that this component doesn't play just like Ingemar's Otello,
  but I really don't know why !

  Paul TOTH <tothpaul@multimania.com>
  http://www.multimania.com/tothpaul
}

{ Orginal version use a MinMax function to play (see otello2.txt)

  With my componen, you have to call TOtello.Thinking each time you want that the component think.
  You can use a timer or just Application.OnIdle !
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

Const
 Depth=6; // Skill:1,2,4,6

type
 // Playing board (with border cells)
  TBoard=array[0..9,0..9] of byte;
 // one main board + thinking boards
  TBoards=array[0..Depth] of TBoard;
 // cells values
  TValues=array[1..8,1..8] of integer;
 // Computer level
  TSkill=(skBeginner,skNovice,skExpert,skMaster);

 // Otello component
  TOtello = class(TCustomControl)
  private
   fComputer:TColor;
   fHuman:TColor;
   fSkill:TSkill;
   fDepth:integer;
   fMoveCount:integer;
   fBoards:TBoards;
   fBoard:integer;
   fWho:integer;
   fKill:integer;
   fThinking:Boolean;  // Computer turn ?
   fCheat:boolean; // Cheat Mode (Game/Hint)
   fScore:integer;
   fThinkMove:integer;
   fCellTop:integer;
   fCellLeft:integer;
   fCellWidth:integer;
   fCellHeight:integer;
   fCellX:integer;
   fCellY:integer;
   fMoveStack:array[0..Depth] of integer;
   fBestMoves:array[0..Depth] of integer;
   fMinMax:array[0..Depth] of record Min,Max:integer end;
   EOnBadMove:TNotifyEvent;
   EOnThink:TNotifyEvent;
   EOnPlay:TNotifyEvent;
   EOnWin:TNotifyEvent;
   EOnLost:TNotifyEvent;
   procedure SetSkill(Value:TSkill);
   procedure ClearBoards;
   function ThisWay(Way:integer):integer;
   function CanPlay(Who,Kill:integer):boolean;
   function CanPlayXY:boolean;
   procedure Eat(x,y:integer);
   procedure Play;
   function PlayerTurn:boolean;
   function CanMove:boolean;
   procedure DoMove;
   procedure Undo;
   procedure TheScoreIs(Value:integer);
   function GetScore(var Values:TValues):integer;
   function CurrentScore:integer;
   function FinalScore(Who,Kill:integer):integer;
   procedure EndOfGame;
  protected
   constructor Create(AOwner:TComponent); override;
   procedure Paint; override;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
   function Thinking:boolean;
   procedure NewGame;
   procedure Pass;
   procedure Cheat;
   property MoveCount:integer read fMoveCount;
   property Score:integer read fScore;
  published
   property Computer:TColor read fComputer write fComputer default clBlue;
   property Human:TColor read fHuman write fHuman default clRed;
   property OnBadMove:TNotifyEvent read EOnBadMove write EOnBadMove;
   property OnThink:TNotifyEvent read EOnThink write EOnThink;
   property OnPlay:TNotifyEvent read EOnPlay write EOnPlay;
   property OnWin:TNotifyEvent read EOnWin write EOnWin;
   property OnLost:TNotifyEvent read EOnLost write EOnLost;
   property Skill:TSkill read fSkill write SetSkill default skBeginner;
   property Color;
   property ParentColor;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MySoft', [TOtello]);
end;

Const
// Computer Skill
 SkillDepth:array[TSkill] of byte=(1,2,4,6);

// movement
 Mvt:array[0..7] of record dx,dy:integer end =(
  (dx:-1;dy:-1),(dx: 0;dy:-1),(dx:+1;dy:-1),
  (dx:-1;dy: 0),              (dx:+1;dy: 0),
  (dx:-1;dy:+1),(dx: 0;dy:+1),(dx:+1;dy:+1)
 );

// Moves Ordered by preference.
 Moves:array[0..59,0..1] of byte =(
  (1,1),(1,8),(8,1),(8,8),  (1,3),(3,1),(1,6),(6,1),
  (3,8),(8,3),(6,8),(8,6),  (1,4),(4,1),(1,5),(5,1),
  (4,8),(8,4),(5,8),(8,5),  (3,3),(3,6),(6,3),(6,6),
  (3,4),(3,5),(4,3),(4,6),  (5,3),(5,6),(6,4),(6,5),
  (2,4),(2,5),(4,2),(4,7),  (5,2),(5,7),(7,4),(7,5),
  (2,3),(2,6),(3,2),(3,7),  (6,2),(6,7),(7,3),(7,6),
  (1,2),(1,7),(2,1),(2,8),  (7,1),(7,8),(8,2),(8,7),
  (2,2),(2,7),(7,2),(7,7)
 );

// Values to compute a "CurrentScore" (a strategic score)
 Value1:TValues=(
  (99, -8, 8, 6, 6, 8, -8,99),
  (-8,-24,-4,-3,-3,-4,-24,-8),
  ( 8, -4, 7, 4, 4, 7, -4, 8),
  ( 6, -3, 4, 0, 0, 4, -3, 6),
  ( 6, -3, 4, 0, 0, 4, -3, 6),
  ( 8, -4, 7, 4, 4, 7, -4, 8),
  (-8,-24,-4,-3,-3,-4,-24,-8),
  (99, -8, 8, 6, 6, 8, -8,99)
 );
 Value2:TValues=(
  (99, -8, 8, 6, 6, 8, -8,99),
  (-8,-24, 0, 1, 1, 0,-24,-8),
  ( 8,  0, 7, 4, 4, 7,  0, 8),
  ( 6,  1, 4, 1, 1, 4,  1, 6),
  ( 6,  1, 4, 1, 1, 4,  1, 6),
  ( 8,  0, 7, 4, 4, 7,  0, 8),
  (-8,-24, 0, 1, 1, 1,-24,-8),
  (99, -8, 8, 6, 6, 8, -8,99)
 );

constructor TOtello.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fComputer:=clBlue;
  fHuman:=clRed;
  Skill:=skBeginner; // set also fDepth
  NewGame;
 end;
 
procedure TOtello.Paint;
 var
  x,y,z:integer;
  ox,oy:integer;
 begin
 // compute cell size
  fCellWidth :=ClientWidth div 8;
  fCellHeight:=ClientHeight div 8;
  if (fCellWidth=0)or(fCellHeight=0) then exit;

  with Canvas do begin
   Brush.Color := Self.Color;
   FillRect(GetClientRect);

  // compute board position
   fCellWidth :=ClientWidth div 8;
   fCellHeight:=ClientHeight div 8;
   fCellLeft:=(ClientWidth-8*fCellWidth) div 2;
   fCellTop:=(ClientHeight-8*fCellHeight) div 2;

  // vertical lines
   ox:=fCellLeft;
   for x:=1 to 9 do begin
    MoveTo(ox,fCellTop); LineTo(ox,ClientHeight-fCellTop);
    inc(ox,fCellWidth);
   end;

  // horizontal lines & moves
   oy:=fCellTop;
   for y:=1 to 8 do begin
    MoveTo(fCellLeft,oy); LineTo(ClientWidth-fCellLeft,oy);
    ox:=fCellLeft;
    for x:=1 to 8 do begin
     z:=fBoards[0][x,y];
     if z>0 then begin
      if z=1 then Brush.Color:=fHuman else Brush.Color:=fComputer;
      Ellipse(ox+2,oy+2,ox+fCellWidth-2,oy+fCellHeight-2);
     end;
     inc(ox,fCellWidth);
    end;
    inc(oy,fCellHeight);
   end;
   MoveTo(fCellLeft,oy); LineTo(ClientWidth-fCellLeft,oy);
  end;
 end;

procedure TOtello.SetSkill(Value:TSkill);
 begin
  fSkill:=Value;
  fDepth:=SkillDepth[fSkill];
 end;

procedure TOtello.ClearBoards;
 var
  i:integer;
 begin
  FillChar(fBoards[0],SizeOf(TBoard),0);
  fBoards[0][4,4]:=1; // Human
  fBoards[0][5,5]:=1;
  fBoards[0][4,5]:=2; // Computer
  fBoards[0][5,4]:=2;
  for i:=0 to 9 do begin
   fBoards[0][0,i]:=3; // border (can't play there)
   fBoards[0][9,i]:=3;
   fBoards[0][i,0]:=3;
   fBoards[0][i,9]:=3;
  end;
 end;

function TOtello.ThisWay(Way:integer):integer;
 var
  dx,dy,ix,iy,b:integer;
 begin
 // compute number of Eaten enemy on "this way"
  Result:=0;
  dx:=mvt[Way].dx; ix:=fCellX+dx;
  dy:=mvt[Way].dy; iy:=fCellY+dy;
  b:=fBoards[fBoard][ix,iy];
  if (b=fKill) then begin
   repeat
    inc(Result);
    inc(ix,dx);
    inc(iy,dy);
    b:=fBoards[fBoard][ix,iy];
   until b<>fKill;
   if b=fWho then exit; // ok
   Result:=0; // not this way
  end;
 end;

function TOtello.CanPlay(Who,Kill:integer):boolean;
 var
  m:integer;
 begin
  Result:=True;
  fWho:=Who;
  fKill:=Kill;
  for m:=0 to 59 do begin // try all moves
   fCellX:=Moves[m,0];
   fCellY:=Moves[m,1];
   if CanPlayXY then exit; // found one possible move
  end;
  Result:=False; // done for this Board.
 end;

Function TOtello.CanPlayXY:boolean;
 var
  m,dx,dy,ix,iy,b:integer;
 begin
  Result:=True;
 // Try to find a way where we Eat enemy
  if (fBoards[fBoard][fCellX,fCellY])=0 then begin
   for m:=0 to 7 do if ThisWay(m)>0 then exit;
  end;
  Result:=False;
 end;

procedure TOtello.Eat(x,y:integer);
 var
  R:TRect;
 begin
  inc(fMoveCount);
  fBoards[fBoard][x,y]:=fWho;
 // Invalidate the Cell on first Board
  if fBoard=0 then begin
   R.Right:=fCellLeft+fCellWidth*x;
   R.Bottom:=fCellTop+fCellHeight*y;
   R.Left:=R.Right-fCellWidth;
   R.Top:=R.Bottom-fCellHeight;
   InvalidateRect(Handle,@R,False);
  end;
 end;

procedure TOtello.Play;
 var
  m,dx,dy,ix,iy,mm:integer;
 begin
  Eat(fCellX,fCellY);
  for m:=0 to 7 do begin
   mm:=ThisWay(m);
   if mm>0 then begin
    dx:=mvt[m].dx; ix:=fCellX+dx;
    dy:=mvt[m].dy; iy:=fCellY+dy;
    while mm>0 do begin
     Eat(ix,iy);
     inc(ix,dx);
     inc(iy,dy);
     dec(mm);
    end;
   end;
  end;
 end;

procedure TOtello.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 begin
  if (fThinking)or(fCellWidth=0)or(fCellHeight=0) then exit;
  SetCaptureControl(Self);
 end;

procedure TOtello.MouseMove(Shift: TShiftState; X, Y: Integer);
 begin
  if fThinking then Cursor:=crAppStart else begin
   if (fCellWidth>0)and(fCellHeight>0) then begin
    fCellX:=1+(X-fCellLeft) div fCellWidth;
    fCellY:=1+(Y-fCellTop) div fCellHeight;
    fWho:=1;
    fKill:=2;
    fBoard:=0;
    if CanPlayXY then Cursor:=crCross else Cursor:=crDefault;
   end;
  end;
 end;

procedure TOtello.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 begin
  if fThinking then exit;
  SetCaptureControl(nil);
  fCellX:=1+(X-fCellLeft) div fCellWidth;
  fCellY:=1+(Y-fCellTop) div fCellHeight;
  fWho:=1; // Human
  fKill:=2; // Computer
  fBoard:=0; // First Level
  if CanPlayXY then begin
   Play;
   Pass;
  end else
   if Assigned(EOnBadMove) then
    EOnBadMove(Self)
   else
    ShowMessage('You may only move to space where the cursor is a cross');
 end;

function TOtello.Thinking:boolean;
 begin
  if fThinking then begin
   if CanMove then DoMove else Undo;
  end;
  Result:=fThinking;
 end;

Procedure TOtello.NewGame;
 begin
  fScore:=0;
  fThinking:=False;
  ClearBoards;
  fMoveCount:=0;
  Invalidate;
 end;

Procedure TOtello.Pass;
 begin
  if CanPlay(2,1) then begin
   fCheat:=False;
   fBestMoves[0]:=60; // none
   fThinkMove:=0;
   fMinMax[0].Min:=-32767;
   fMinMax[0].Max:=+32767;
   fThinking:=True;
   if Assigned(EOnThink) then EOnThink(Self);
  end else begin
   if not CanPlay(1,2) then EndOfGame;
  end;
 end;

procedure TOtello.Cheat;
 begin
  fCheat:=True;
  fBestMoves[0]:=60; // none
  fThinkMove:=0;
  fMinMax[0].Min:=-32767;
  fMinMax[0].Max:=+32767;
  fThinking:=True;
 end;

function TOtello.PlayerTurn:boolean;
 begin
  Result:=odd(fBoard);
  if fCheat then Result:=Not Result;
 end;

function TOtello.CanMove:boolean;
 begin
  Result:=True;
  if PlayerTurn then begin // player turn
   fWho:=1; fKill:=2;
  end else begin // computer turn
   fWho:=2; fKill:=1;
  end;
  while fThinkMove<60 do begin // try another move
   fCellX:=Moves[fThinkMove,0];
   fCellY:=Moves[fThinkMove,1];
   inc(fThinkMove);
   if CanPlayXY then exit; // found one possible move
  end;
  Result:=False; // done for this Board.
 end;

procedure TOtello.DoMove;
 var
  m,mm:integer;
  ix,iy,dx,dy:integer;
  s:integer;
 begin
  if fBoard>=fDepth then begin // we've reach Depht thinking (but still can move)
   if PlayerTurn then begin // Computer turn (for CurrentScore)
    fWho:=2; fKill:=1;
   end else begin // Player turn
    fWho:=1; fKill:=2;
   end;
   TheScoreIs(CurrentScore);
   exit;
  end;
  // save ThinkMove & start a new loop
  fMoveStack[fBoard]:=fThinkMove;
  fThinkMove:=0;
  // copy current board
  Move(fBoards[fBoard],fBoards[fBoard+1],SizeOf(TBoard));
  // MinMax
  with fMinMax[fBoard+1] do begin
   Min:=-fMinMax[fBoard].Max;
   Max:=-fMinMax[fBoard].Min;
  end;
  // next board
  inc(fBoard);
  // move
  fBoards[fBoard][fCellX,fCellY]:=fWho;
  for m:=0 to 7 do begin
   mm:=ThisWay(m);
   if mm>0 then begin
    dx:=mvt[m].dx; ix:=fCellX+dx;
    dy:=mvt[m].dy; iy:=fCellY+dy;
    while mm>0 do begin
     fBoards[fBoard][ix,iy]:=fWho;
     inc(ix,dx);
     inc(iy,dy);
     dec(mm);
    end;
   end;
  end;
 end;

procedure TOtello.Undo;
 var
  Value:integer;
  w,k:integer;
  p:TPoint;
 begin
  if fBoard=0 then begin // done with all possible moves
   fThinking:=False;
   Value:=fBestMoves[0];
   if Value<60 then begin // have we found a possible move ?
    fCellX:=Moves[Value,0];
    fCellY:=Moves[Value,1];
    if fCheat then begin
     p.x:=fCellLeft+fCellWidth *fCellX-fCellWidth div 2;
     p.y:=fCellTop +fCellHeight*fCellY-fCellHeight div 2;
     p:=ClientToScreen(p);
     SetCursorPos(p.x,p.y);
    end else begin
     Play;
     if not CanPlay(1,2) then begin
      Pass;
     end else begin
      if Assigned(EOnPlay) then EOnPlay(Self);
     end;
    end;
   end else begin
    ShowMessage('Unable to move :(');
   end;
  end else begin
  // We Can't move...check for enemy move
   if CanPlay(fKill,fWho) then
    TheScoreIs(CurrentScore)
   else
    TheScoreIs(FinalScore(fWho,fKill)); // CanPlay() reset fWho & fKill
  end;
 end;

procedure TOtello.TheScoreIs(Value:integer);
 begin
  dec(fBoard);
  fThinkMove:=fMoveStack[fBoard];
  if Value>fMinMax[fBoard].Min then begin
   fMinMax[fBoard].Min:=Value;
   fBestMoves[fBoard]:=fThinkMove-1; // "-1" cause CanMove inc(fThinkMove)
   if Value>=fMinMax[fBoard].Max then begin
    if fBoard=0 then Undo else TheScoreIs(-Value);
   end;
  end;
 end;

function TOtello.CurrentScore:integer;
 var
  b11,b18,b81,b88:byte;
 begin
   b11 := fBoards[fBoard][1,1];
   b18 := fBoards[fBoard][1,8];
   b81 := fBoards[fBoard][8,1];
   b88 := fBoards[fBoard][8,8];
   if (b11>0)or(b18>0)or(b81>0)or(b88>0) then begin
    if b11=0 then begin
     value2[1,2]:=-8;
     value2[2,1]:=-8;
     value2[2,2]:=-24;
    end else begin
     value2[1,2]:=12;
     value2[2,1]:=12;
     value2[2,2]:=8;
    end;
    if b81=0 then begin
     value2[8,2]:=-8;
     value2[7,1]:=-8;
     value2[7,2]:=-24;
    end else begin
     value2[8,2]:=12;
     value2[7,1]:=12;
     value2[7,2]:=8;
    end;
    if b88=0 then begin
     value2[8,7]:=-8;
     value2[7,8]:=-8;
     value2[7,7]:=-24;
    end else begin
     value2[8,7]:=12;
     value2[7,8]:=12;
     value2[7,7]:=8;
    end;
    result:=GetScore(Value2);
   end else begin
    result:=GetScore(Value1);
   end;
 end;

function TOtello.GetScore(Var Values:TValues):integer;
 var
  x,y,z,v:integer;
  sWho,sKill:integer;
  bKill:boolean;
 begin
  sWho:=0;
  sKill:=0;
  bKill:=false;
  for x:=1 to 8 do
   for y:=1 to 8 do begin
    z:=fBoards[fBoard][x,y];
    v:=Values[x,y];
    if z=fWho then inc(sWho,v) else
    if z=fKill then begin inc(sKill,v); bKill:=true end;
   end;
  if bKill then Result:=sWho-sKill else Result:=32000;
 end;

function TOtello.FinalScore(Who,Kill:integer):integer;
 var
  x,y,z:integer;
 begin
  Result:=0;
  for x:=1 to 8 do
   for y:=1 to 8 do begin
    z:=fBoards[fBoard][x,y];
    if z=0 then begin
     fWho:=Who;
     Eat(x,y);
     z:=Who;
    end;
    if z=Who then inc(Result) else
    if z=Kill then dec(Result);
   end;
  if Result>0 then inc(Result,32000) else
  if Result<0 then dec(Result,32000);
 end;

Procedure TOtello.EndOfGame;
 begin
  fScore:=FinalScore(1,2);
  if fScore>=0 then begin
   dec(fScore,32000);
   if Assigned(EOnWin) then EOnWin(Self) else ShowMessage('You win by '+IntToStr(fScore))
  end else begin
   inc(fScore,32000);
   if Assigned(EOnLost) then EOnLost(Self) else ShowMessage('You lost by : '+IntToStr(fScore));
  end;
 end;

end.


